/*    hivelyplayer.cpp
 *
 * Hively player
 *
 * Copyright 2008 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#include "hivelyplayer.h"

const TInt KSampleRate = 44100;
const TInt KHivelyRenderSamples = KSampleRate/50;
const TInt KBufferSamples = KSampleRate/10; // must be >= KHivelyRenderSamples
const TInt KNumBuffers = 8;

_LIT(KAssertPanic, "hivelyplayer.cpp");
#define assert(x) __ASSERT_ALWAYS(x, User::Panic(KAssertPanic, __LINE__))



// Factory
CHivelyPlayer *CHivelyPlayer::NewL(MHivelyPlayerObserver &aObserver) {
    CHivelyPlayer *self = new (ELeave) CHivelyPlayer(aObserver);
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop();
    return self;
}


// Constructor
CHivelyPlayer::CHivelyPlayer(MHivelyPlayerObserver &aObserver)
    : iObserver(aObserver) {
    iVolume = 50;
}


// Destructor
CHivelyPlayer::~CHivelyPlayer() {
    Stop();
    delete [] iLeftBuffer;
    delete [] iRightBuffer;
    for ( TInt i = 0; i < KNumBuffers; i++ ) {
        delete [] iOutputBuffers[i];
    }
    delete [] iOutputBuffers;
    delete [] iOutputPtrs;
}


// Second-phase constructor
void CHivelyPlayer::ConstructL() {

    hvl_InitReplayer();

    // Allocate buffers
    iLeftBuffer = new (ELeave) TInt16[KHivelyRenderSamples];
    iRightBuffer = new (ELeave) TInt16[KHivelyRenderSamples];
    iOutputBuffers = new (ELeave) TInt16*[KNumBuffers];
    Mem::FillZ(iOutputBuffers, KNumBuffers*sizeof(TInt16));
    iOutputPtrs = new (ELeave) TPtrC8[KNumBuffers];
    for ( TInt i = 0; i < KNumBuffers; i++ ) {
        iOutputBuffers[i] = new (ELeave) TInt16[2*KBufferSamples];
        iOutputPtrs[i].Set((TUint8*)iOutputBuffers[i], 4*KBufferSamples);
    }
}


// Start playback
void CHivelyPlayer::PlayL(const TDesC &aFileName, RFs & /*aFs*/) {
    
    // Convert name to ASCIIZ (and hope it's all ASCII)
    HBufC8 *buf = HBufC8::NewLC(aFileName.Length()+1);
    TPtr8 ptr = buf->Des();
    ptr.Copy(aFileName);
    ptr.Append('\0');

    // Load and initialize the tune
    iTune = hvl_LoadTune((TEXT*) ptr.Ptr(), KSampleRate, 4);
    if ( iTune == NULL ) {
        User::Leave(KErrGeneral);
    }
    hvl_InitSubsong(iTune, 0);
    
    CleanupStack::PopAndDestroy(); //buf

    // Create and initialize an audio output stream
    iStream = CMdaAudioOutputStream::NewL(*this);
    iStream->Open(&iSettings);
}


// Stop playback
void CHivelyPlayer::Stop() {
    if ( iTune ) {
        hvl_FreeTune(iTune);
        iTune = NULL;
    }
    delete iStream;
    iStream = NULL;
    iStreamInit = EFalse;
}


// Set volume
void CHivelyPlayer::SetVolume(TInt aVolume) {
    iVolume = aVolume;
    if ( iStreamInit ) {
        iStream->SetVolume((iVolume * iStream->MaxVolume()) / 100);
    }
}


// Get title
HBufC *CHivelyPlayer::TitleLC() {
    if ( !iTune ) {
        return HBufC::NewLC(0);
    }

    TInt len = 0;
    while ( len < 128 ) {
        if ( iTune->ht_Name[len] == 0 ) {
            break;
        } else {
            len++;
        }
    }

    HBufC *buf = HBufC::NewLC(len);
    TPtr ptr = buf->Des();
    ptr.Copy(TPtrC8((TUint8*)iTune->ht_Name, len));
    return buf;
}


// Play some data
void CHivelyPlayer::WriteSoundL(TInt aBuffer) {

    // The Hively replay outputs separate left and right buffer, while
    // CMdaAudioOutputStream needs them interleaved. Therefore we must
    // interleave the Hively output before playing it out.

    // Since the Hively replay will happily output 1/50s worth of data forever,
    // we'll always render and output a whole output buffer.

    // First, let's interleave any leftover data from previous round
    TInt left = KBufferSamples;
    TInt16 *out = iOutputBuffers[aBuffer];
    while ( (left > 0) && (iHivelyBufPos < KHivelyRenderSamples) ) {
        *out++ = iLeftBuffer[iHivelyBufPos];
        *out++ = iRightBuffer[iHivelyBufPos];
        iHivelyBufPos++;
        left--;
    }

    // Now render more data and interleave it until the output buffer is full
    while ( left > 0 ) {
        hvl_DecodeFrame(iTune, (int8*) iLeftBuffer, (int8*) iRightBuffer, 2);
        iHivelyBufPos = 0;
        while ( (left > 0) && (iHivelyBufPos < KHivelyRenderSamples) ) {
            *out++ = iLeftBuffer[iHivelyBufPos];
            *out++ = iRightBuffer[iHivelyBufPos];
            iHivelyBufPos++;
            left--;
        }        
    }

    // Send the data to the output stream
    iStream->WriteL(iOutputPtrs[aBuffer]);
}


// MMdaAudioOutputStreamCallback::MaoscOpenComplete
void CHivelyPlayer::MaoscOpenComplete(TInt aError) {
    if ( aError != KErrNone ) {
        iObserver.MhpoAudioError(aError);
        return;
    }

    iStream->SetAudioPropertiesL(TMdaAudioDataSettings::ESampleRate44100Hz,
                                 TMdaAudioDataSettings::EChannelsStereo);
    iStream->SetVolume((iVolume * iStream->MaxVolume()) / 100);
    iStreamInit = ETrue;

    // Fill all buffers and play them
    for ( TInt i = 0; i < KNumBuffers; i++ ) {
        TRAPD(error, WriteSoundL(i));
        if ( error != KErrNone ) {
            iObserver.MhpoAudioError(aError);
            return;
        }
    }
}


// MMdaAudioOutputStreamCallback::MaoscBufferCopied
void CHivelyPlayer::MaoscBufferCopied(TInt aError, const TDesC8 &aBuffer) {
    if ( aError == KErrUnderflow ) {
        iObserver.MhpoUnderflow();
    } else if ( aError != KErrNone ) {
        iObserver.MhpoAudioError(aError);
        return;
    }

    // Figure out which buffer became free, fill it, and play it
    TInt i;
    for ( i = 0; i < KNumBuffers; i++ ) {
        if ( aBuffer.Ptr() == iOutputPtrs[i].Ptr() ) {
            TRAPD(error, WriteSoundL(i));
            if ( error != KErrNone ) {
                iObserver.MhpoAudioError(aError);
                return;
            }
            break;
        }
    }
    assert(i < KNumBuffers);
}


// MMdaAudioOutputStreamCallback::MaoscPlayComplete
void CHivelyPlayer::MaoscPlayComplete(TInt aError) {
    if ( aError == KErrUnderflow ) {
        iObserver.MhpoUnderflow();
    } else if ( aError != KErrNone ) {
        iObserver.MhpoAudioError(aError);
        return;
    }
}
