/*    hivelyplayer.h
 *
 * Hively player
 *
 * Copyright 2008 Petteri Kangaslampi
 *
 * See license.txt for full copyright and license information.
*/

#ifndef HIVELYPLAYER_H
#define HIVELYPLAYER_H

#include <e32base.h>
#include <mdaaudiooutputstream.h>
#include <mda/common/audio.h>

extern "C" {
#include "types.h"
#include "hvl_replay.h"
}


class RFs;


/**
 * A hively player observer class, implemented by clients
 */
class MHivelyPlayerObserver {

public:
    /** 
     * An audio error has occurred and playback has been stopped. The client
     * should reinitialize the player.
     * 
     * @param aError Error code
     */
    virtual void MhpoAudioError(TInt aError) = 0;

    /** 
     * An underflow error has occurred, but playback still proceeds.
     * 
     */
    virtual void MhpoUnderflow() = 0;
};


/**
 * A hively player wrapper class.
 */
class CHivelyPlayer : public CBase, MMdaAudioOutputStreamCallback {

public:
    /** 
     * Factory method, creates a new hively player instance
     * 
     * @param aObserver Observer object
     * 
     * @return A new CHivelyPlayer instance
     */
    static CHivelyPlayer *NewL(MHivelyPlayerObserver &aObserver);

    /** 
     * Destructor.
     */
    ~CHivelyPlayer();

    /** 
     * Starts playing a new file
     * 
     * @param aFileName The file to play
     * @param aFs File server session to use, usually from CCoeEnv
     */
    void PlayL(const TDesC &aFileName, RFs &aFs);
    
    /** 
     * Stops playback.
     */
    void Stop();

    /** 
     * Set volume
     * 
     * @param aVolume Volume, in percentage (0 = min, 100 = max)
     */
    void SetVolume(TInt aVolume);

    /** 
     * Gets the song title.
     * 
     * @return Buffer descriptor containing the title. The buffer is left
     *         on the cleanup stack. Empty if no tune is loaded.
     */    
    HBufC *TitleLC();

private:
    CHivelyPlayer(MHivelyPlayerObserver &aObserver);
    void ConstructL();
    void WriteSoundL(TInt aBuffer);

private: // MMdaAudioOutputStreamCallback
    void MaoscOpenComplete(TInt aError);
    void MaoscBufferCopied(TInt aError, const TDesC8 &aBuffer);
    void MaoscPlayComplete(TInt aError);

    MHivelyPlayerObserver &iObserver;
    struct hvl_tune *iTune;
    CMdaAudioOutputStream *iStream;
    TMdaAudioDataSettings iSettings; // dummy
    TInt16 *iLeftBuffer;
    TInt16 *iRightBuffer;
    TInt iHivelyBufPos;
    TInt16 **iOutputBuffers;
    TPtrC8 *iOutputPtrs;
    TInt iVolume;
    TBool iStreamInit;
};


#endif
